Package: brms
Encoding: UTF-8
Type: Package
Title: Bayesian Regression Models using 'Stan'
Version: 2.22.0
Date: 2024-09-20
Authors@R: 
    c(person("Paul-Christian", "Bürkner", email = "paul.buerkner@gmail.com",
             role = c("aut", "cre")),
      person("Jonah", "Gabry", role = c("ctb")),
      person("Sebastian", "Weber", role = c("ctb")),
      person("Andrew", "Johnson", role = c("ctb")),
      person("Martin", "Modrak", role = c("ctb")),
      person("Hamada S.", "Badr", role = c("ctb")),
      person("Frank", "Weber", role = c("ctb")),
      person("Aki", "Vehtari", role = c("ctb")),
      person("Mattan S.", "Ben-Shachar", role = c("ctb")),
      person("Hayden", "Rabel", role = c("ctb")),
      person("Simon C.", "Mills", role = c("ctb")),
      person("Stephen", "Wild", role = c("ctb")),
      person("Ven", "Popov", role = c("ctb")))
Depends: R (>= 3.6.0), Rcpp (>= 0.12.0), methods
Imports: rstan (>= 2.29.0), ggplot2 (>= 2.0.0), loo (>= 2.8.0),
        posterior (>= 1.6.0), Matrix (>= 1.1.1), mgcv (>= 1.8-13),
        rstantools (>= 2.1.1), bayesplot (>= 1.5.0), bridgesampling (>=
        0.3-0), glue (>= 1.3.0), rlang (>= 1.0.0), future (>= 1.19.0),
        future.apply (>= 1.0.0), matrixStats, nleqslv, nlme, coda,
        abind, stats, utils, parallel, grDevices, backports
Suggests: testthat (>= 0.9.1), emmeans (>= 1.4.2), cmdstanr (>= 0.5.0),
        projpred (>= 2.0.0), priorsense (>= 1.0.0), shinystan (>=
        2.4.0), splines2 (>= 0.5.0), RWiener, rtdists, extraDistr,
        processx, mice, spdep, mnormt, lme4, MCMCglmm, ape, arm,
        statmod, digest, diffobj, R.rsp, gtable, shiny, knitr,
        rmarkdown
Description: Fit Bayesian generalized (non-)linear multivariate multilevel models
    using 'Stan' for full Bayesian inference. A wide range of distributions
    and link functions are supported, allowing users to fit -- among others --
    linear, robust linear, count data, survival, response times, ordinal,
    zero-inflated, hurdle, and even self-defined mixture models all in a
    multilevel context. Further modeling options include both theory-driven and
    data-driven non-linear terms, auto-correlation structures, censoring and 
    truncation, meta-analytic standard errors, and quite a few more. 
    In addition, all parameters of the response distribution can be predicted 
    in order to perform distributional regression. Prior specifications are 
    flexible and explicitly encourage users to apply prior distributions that
    actually reflect their prior knowledge. Models can easily be evaluated and
    compared using several methods assessing posterior or prior predictions. 
    References: Bürkner (2017) <doi:10.18637/jss.v080.i01>; 
    Bürkner (2018) <doi:10.32614/RJ-2018-017>;
    Bürkner (2021) <doi:10.18637/jss.v100.i05>; Carpenter et al. (2017)
    <doi:10.18637/jss.v076.i01>.
LazyData: true
NeedsCompilation: no
License: GPL-2
URL: https://github.com/paul-buerkner/brms,
        https://discourse.mc-stan.org/, https://paulbuerkner.com/brms/
BugReports: https://github.com/paul-buerkner/brms/issues
Additional_repositories: https://stan-dev.r-universe.dev/
VignetteBuilder: knitr, R.rsp
RoxygenNote: 7.3.2
Packaged: 2024-09-23 05:27:41 UTC; paul-buerkner
Author: Paul-Christian Bürkner [aut, cre],
  Jonah Gabry [ctb],
  Sebastian Weber [ctb],
  Andrew Johnson [ctb],
  Martin Modrak [ctb],
  Hamada S. Badr [ctb],
  Frank Weber [ctb],
  Aki Vehtari [ctb],
  Mattan S. Ben-Shachar [ctb],
  Hayden Rabel [ctb],
  Simon C. Mills [ctb],
  Stephen Wild [ctb],
  Ven Popov [ctb]
Maintainer: Paul-Christian Bürkner <paul.buerkner@gmail.com>
Repository: CRAN
Date/Publication: 2024-09-23 13:00:29 UTC
Built: R 4.3.3; ; 2024-09-23 15:33:49 UTC; unix
