# Generated by roxygen2: do not edit by hand

S3method("+",bform)
S3method("+",brmsprior)
S3method("+",stanvars)
S3method(.compute_point_estimate,brmsprep)
S3method(.compute_point_estimate,mvbrmsprep)
S3method(.extract_par_terms,brmsfit)
S3method(.extract_par_terms,brmsterms)
S3method(.extract_par_terms,mvbrmsterms)
S3method(.frame_index,brmsterms)
S3method(.frame_index,mvbrmsterms)
S3method(.ndraws,brmsprep)
S3method(.ndraws,mvbrmsprep)
S3method(.thin_draws,brmsprep)
S3method(.thin_draws,mvbrmsprep)
S3method(LOO,brmsfit)
S3method(VarCorr,brmsfit)
S3method(WAIC,brmsfit)
S3method(add_criterion,brmsfit)
S3method(add_ic,brmsfit)
S3method(as.array,brmsfit)
S3method(as.data.frame,brmsfit)
S3method(as.matrix,brmsfit)
S3method(as.mcmc,brmsfit)
S3method(as_draws,brmsfit)
S3method(as_draws_array,brmsfit)
S3method(as_draws_df,brmsfit)
S3method(as_draws_list,brmsfit)
S3method(as_draws_matrix,brmsfit)
S3method(as_draws_rvars,brmsfit)
S3method(autocor,brmsfit)
S3method(bayes_R2,brmsfit)
S3method(bayes_factor,brmsfit)
S3method(bridge_sampler,brmsfit)
S3method(brmsframe,brmsterms)
S3method(brmsframe,btl)
S3method(brmsframe,btnl)
S3method(brmsframe,default)
S3method(brmsframe,mvbrmsterms)
S3method(brmsterms,brmsformula)
S3method(brmsterms,default)
S3method(brmsterms,mvbrmsformula)
S3method(c,brmsprior)
S3method(c,stanvars)
S3method(coef,brmsfit)
S3method(compute_xi,brmsfit)
S3method(compute_xi,brmsprep)
S3method(compute_xi,mvbrmsprep)
S3method(conditional_effects,brmsfit)
S3method(conditional_effects,brmsterms)
S3method(conditional_effects,mvbrmsterms)
S3method(conditional_smooths,brmsfit)
S3method(conditional_smooths,brmsterms)
S3method(conditional_smooths,btl)
S3method(conditional_smooths,default)
S3method(conditional_smooths,mvbrmsterms)
S3method(control_params,brmsfit)
S3method(data_predictor,brmsterms)
S3method(data_predictor,btl)
S3method(data_predictor,btnl)
S3method(data_predictor,mvbrmsterms)
S3method(data_response,brmsframe)
S3method(data_response,mvbrmsframe)
S3method(def_scale_prior,brmsterms)
S3method(def_scale_prior,mvbrmsframe)
S3method(default_prior,brmsfit)
S3method(default_prior,default)
S3method(dpar_family,default)
S3method(dpar_family,mixfamily)
S3method(duplicated,brmsprior)
S3method(exclude_pars,bframel)
S3method(exclude_pars,brmsfit)
S3method(exclude_pars,brmsframe)
S3method(exclude_pars,default)
S3method(exclude_pars,mvbrmsframe)
S3method(exclude_terms,brmsfit)
S3method(exclude_terms,brmsformula)
S3method(exclude_terms,mvbrmsformula)
S3method(expose_functions,brmsfit)
S3method(family,brmsfit)
S3method(family_bounds,brmsterms)
S3method(family_bounds,mvbrmsterms)
S3method(family_info,"NULL")
S3method(family_info,brmsfamily)
S3method(family_info,brmsfit)
S3method(family_info,brmsformula)
S3method(family_info,brmsterms)
S3method(family_info,btl)
S3method(family_info,btnl)
S3method(family_info,default)
S3method(family_info,family)
S3method(family_info,list)
S3method(family_info,mixfamily)
S3method(family_info,mvbrmsformula)
S3method(family_info,mvbrmsterms)
S3method(fitted,brmsfit)
S3method(fixef,brmsfit)
S3method(formula,brmsfit)
S3method(frame_ac,"NULL")
S3method(frame_ac,acframe)
S3method(frame_ac,brmsterms)
S3method(frame_ac,btl)
S3method(frame_ac,btnl)
S3method(frame_ac,default)
S3method(frame_ac,mvbrmsterms)
S3method(frame_basis,brmsterms)
S3method(frame_basis,btl)
S3method(frame_basis,btnl)
S3method(frame_basis,default)
S3method(frame_basis,mvbrmsterms)
S3method(getCall,brmsfit)
S3method(get_ad_vars,brmsterms)
S3method(get_ad_vars,mvbrmsterms)
S3method(get_all_effects,brmsterms)
S3method(get_all_effects,btl)
S3method(get_all_effects,btnl)
S3method(get_all_effects,default)
S3method(get_all_effects,mvbrmsterms)
S3method(get_data2_autocor,brmsformula)
S3method(get_data2_autocor,brmsterms)
S3method(get_data2_autocor,mvbrmsformula)
S3method(get_data2_autocor,mvbrmsterms)
S3method(get_effect,brmsfit)
S3method(get_effect,brmsformula)
S3method(get_effect,brmsterms)
S3method(get_effect,btl)
S3method(get_effect,btnl)
S3method(get_effect,default)
S3method(get_effect,mvbrmsformula)
S3method(get_effect,mvbrmsterms)
S3method(get_element,default)
S3method(get_element,mvbrmsformula)
S3method(get_element,mvbrmsterms)
S3method(get_group_vars,brmsfit)
S3method(get_group_vars,brmsterms)
S3method(get_group_vars,default)
S3method(get_group_vars,mvbrmsterms)
S3method(get_int_vars,brmsterms)
S3method(get_int_vars,mvbrmsterms)
S3method(get_levels,brmsterms)
S3method(get_levels,default)
S3method(get_levels,list)
S3method(get_levels,mvbrmsterms)
S3method(get_re,brmsterms)
S3method(get_re,btl)
S3method(get_re,default)
S3method(get_re,mvbrmsterms)
S3method(get_unused_arg_vars,brmsformula)
S3method(get_unused_arg_vars,brmsterms)
S3method(get_unused_arg_vars,mvbrmsformula)
S3method(get_unused_arg_vars,mvbrmsterms)
S3method(hypothesis,brmsfit)
S3method(hypothesis,default)
S3method(kfold,brmsfit)
S3method(logLik,brmsfit)
S3method(log_lik,brmsfit)
S3method(log_lik,brmsprep)
S3method(log_lik,mvbrmsprep)
S3method(log_posterior,brmsfit)
S3method(loo,brmsfit)
S3method(loo_R2,brmsfit)
S3method(loo_compare,brmsfit)
S3method(loo_epred,brmsfit)
S3method(loo_linpred,brmsfit)
S3method(loo_model_weights,brmsfit)
S3method(loo_moment_match,brmsfit)
S3method(loo_moment_match,loo)
S3method(loo_predict,brmsfit)
S3method(loo_predictive_interval,brmsfit)
S3method(loo_subsample,brmsfit)
S3method(marginal_effects,brmsfit)
S3method(marginal_smooths,brmsfit)
S3method(mcmc_plot,brmsfit)
S3method(model.frame,brmsfit)
S3method(model_weights,brmsfit)
S3method(nchains,brmsfit)
S3method(ndraws,brmsfit)
S3method(neff_ratio,brmsfit)
S3method(ngrps,brmsfit)
S3method(niterations,brmsfit)
S3method(nobs,brmsfit)
S3method(nsamples,brmsfit)
S3method(nuts_params,brmsfit)
S3method(nvariables,brmsfit)
S3method(pairs,brmsfit)
S3method(parnames,brmsfit)
S3method(parnames,default)
S3method(plot,brmsMarginalEffects)
S3method(plot,brms_conditional_effects)
S3method(plot,brmsfit)
S3method(plot,brmshypothesis)
S3method(post_prob,brmsfit)
S3method(posterior_average,brmsfit)
S3method(posterior_epred,brmsfit)
S3method(posterior_epred,brmsprep)
S3method(posterior_epred,mvbrmsprep)
S3method(posterior_interval,brmsfit)
S3method(posterior_linpred,brmsfit)
S3method(posterior_predict,brmsfit)
S3method(posterior_predict,brmsprep)
S3method(posterior_predict,mvbrmsprep)
S3method(posterior_samples,brmsfit)
S3method(posterior_samples,default)
S3method(posterior_smooths,brmsfit)
S3method(posterior_smooths,btl)
S3method(posterior_smooths,btnl)
S3method(posterior_summary,brmsfit)
S3method(posterior_summary,default)
S3method(pp_average,brmsfit)
S3method(pp_check,brmsfit)
S3method(pp_mixture,brmsfit)
S3method(predict,brmsfit)
S3method(predictive_error,brmsfit)
S3method(predictive_interval,brmsfit)
S3method(predictor,bprepl)
S3method(predictor,bprepnl)
S3method(prepare_predictions,bframel)
S3method(prepare_predictions,bframenl)
S3method(prepare_predictions,brmsfit)
S3method(prepare_predictions,brmsframe)
S3method(prepare_predictions,default)
S3method(prepare_predictions,mvbrmsframe)
S3method(print,brmsMarginalEffects)
S3method(print,brms_conditional_effects)
S3method(print,brmsfamily)
S3method(print,brmsfit)
S3method(print,brmsformula)
S3method(print,brmshypothesis)
S3method(print,brmsmodel)
S3method(print,brmsprior)
S3method(print,brmssummary)
S3method(print,cor_arma)
S3method(print,cor_brms_formula)
S3method(print,cor_car)
S3method(print,cor_cosy)
S3method(print,cor_empty)
S3method(print,cor_fixed)
S3method(print,cor_sar)
S3method(print,cov_fixed)
S3method(print,customfamily)
S3method(print,iclist)
S3method(print,loolist)
S3method(print,mixfamily)
S3method(print,mvbrmsformula)
S3method(prior_draws,brmsfit)
S3method(prior_draws,default)
S3method(prior_predictor,bframel)
S3method(prior_predictor,bframenl)
S3method(prior_predictor,brmsframe)
S3method(prior_predictor,default)
S3method(prior_predictor,mvbrmsframe)
S3method(prior_summary,brmsfit)
S3method(priorsense::create_priorsense_data,brmsfit)
S3method(priorsense::log_lik_draws,brmsfit)
S3method(priorsense::log_prior_draws,brmsfit)
S3method(projpred::get_refmodel,brmsfit)
S3method(psis,brmsfit)
S3method(r_eff_log_lik,"function")
S3method(r_eff_log_lik,matrix)
S3method(ranef,brmsfit)
S3method(reloo,brmsfit)
S3method(reloo,loo)
S3method(rename_predictor,bframel)
S3method(rename_predictor,brmsterms)
S3method(rename_predictor,default)
S3method(rename_predictor,mvbrmsterms)
S3method(rescale_old_mo,brmsfit)
S3method(rescale_old_mo,brmsterms)
S3method(rescale_old_mo,btl)
S3method(rescale_old_mo,btnl)
S3method(rescale_old_mo,mvbrmsterms)
S3method(residuals,brmsfit)
S3method(restructure,brmsfit)
S3method(rhat,brmsfit)
S3method(shinystan::launch_shinystan,brmsfit)
S3method(stan_log_lik,brmsterms)
S3method(stan_log_lik,mvbrmsterms)
S3method(stan_predictor,bframel)
S3method(stan_predictor,bframenl)
S3method(stan_predictor,brmsframe)
S3method(stan_predictor,mvbrmsframe)
S3method(stancode,brmsfit)
S3method(stancode,default)
S3method(standata,brmsfit)
S3method(standata,default)
S3method(stanplot,brmsfit)
S3method(subset,psis)
S3method(summarise_families,brmsformula)
S3method(summarise_families,mvbrmsformula)
S3method(summarise_links,brmsformula)
S3method(summarise_links,mvbrmsformula)
S3method(summary,brmsfit)
S3method(summary,customfamily)
S3method(summary,family)
S3method(summary,mixfamily)
S3method(unclass_draws,default)
S3method(unclass_draws,draws_df)
S3method(update,brmsfit)
S3method(update,brmsfit_multiple)
S3method(update,brmsformula)
S3method(update,mvbrmsformula)
S3method(update_old_family,brmsfamily)
S3method(update_old_family,brmsformula)
S3method(update_old_family,customfamily)
S3method(update_old_family,default)
S3method(update_old_family,mixfamily)
S3method(update_old_family,mvbrmsformula)
S3method(update_re_terms,brmsformula)
S3method(update_re_terms,formula)
S3method(update_re_terms,mvbrmsformula)
S3method(valid_dpars,brmsfit)
S3method(valid_dpars,brmsformula)
S3method(valid_dpars,brmsterms)
S3method(valid_dpars,default)
S3method(valid_dpars,mixfamily)
S3method(valid_dpars,mvbrmsformula)
S3method(valid_dpars,mvbrmsterms)
S3method(validate_formula,brmsformula)
S3method(validate_formula,default)
S3method(validate_formula,mvbrmsformula)
S3method(validate_special_prior,brmsprior)
S3method(validate_special_prior,brmsterms)
S3method(validate_special_prior,btl)
S3method(validate_special_prior,btnl)
S3method(validate_special_prior,default)
S3method(validate_special_prior,mvbrmsterms)
S3method(variables,brmsfit)
S3method(vars_keep_na,brmsterms)
S3method(vars_keep_na,mvbrmsterms)
S3method(vcov,brmsfit)
S3method(waic,brmsfit)
export("add_ic<-")
export(Beta)
export(LOO)
export(R2D2)
export(VarCorr)
export(WAIC)
export(acat)
export(acformula)
export(add_criterion)
export(add_ic)
export(add_loo)
export(add_rstan_model)
export(add_waic)
export(ar)
export(arma)
export(as.brmsprior)
export(as.mcmc)
export(as_draws)
export(as_draws_array)
export(as_draws_df)
export(as_draws_list)
export(as_draws_matrix)
export(as_draws_rvars)
export(asym_laplace)
export(autocor)
export(bayes_R2)
export(bayes_factor)
export(bernoulli)
export(beta_binomial)
export(bf)
export(bridge_sampler)
export(brm)
export(brm_multiple)
export(brmsfamily)
export(brmsfit_needs_refit)
export(brmsformula)
export(brmsterms)
export(car)
export(categorical)
export(combine_models)
export(compare_ic)
export(conditional_effects)
export(conditional_smooths)
export(constant)
export(control_params)
export(cor_ar)
export(cor_arma)
export(cor_arr)
export(cor_bsts)
export(cor_car)
export(cor_cosy)
export(cor_errorsar)
export(cor_fixed)
export(cor_icar)
export(cor_lagsar)
export(cor_ma)
export(cor_sar)
export(cosy)
export(cox)
export(cratio)
export(cs)
export(cse)
export(cumulative)
export(custom_family)
export(dasym_laplace)
export(data_predictor)
export(data_response)
export(dbeta_binomial)
export(ddirichlet)
export(default_prior)
export(density_ratio)
export(dexgaussian)
export(dfrechet)
export(dgen_extreme_value)
export(dhurdle_gamma)
export(dhurdle_lognormal)
export(dhurdle_negbinomial)
export(dhurdle_poisson)
export(dinv_gaussian)
export(dirichlet)
export(dlogistic_normal)
export(dmulti_normal)
export(dmulti_student_t)
export(do_call)
export(dshifted_lnorm)
export(dskew_normal)
export(dstudent_t)
export(dvon_mises)
export(dwiener)
export(dzero_inflated_beta)
export(dzero_inflated_beta_binomial)
export(dzero_inflated_binomial)
export(dzero_inflated_negbinomial)
export(dzero_inflated_poisson)
export(empty_prior)
export(exgaussian)
export(exponential)
export(expose_functions)
export(expp1)
export(extract_draws)
export(fcor)
export(fixef)
export(frechet)
export(gen_extreme_value)
export(geometric)
export(get_dpar)
export(get_prior)
export(get_y)
export(gp)
export(gr)
export(horseshoe)
export(hurdle_cumulative)
export(hurdle_gamma)
export(hurdle_lognormal)
export(hurdle_negbinomial)
export(hurdle_poisson)
export(hypothesis)
export(inv_logit_scaled)
export(is.brmsfit)
export(is.brmsfit_multiple)
export(is.brmsformula)
export(is.brmsprior)
export(is.brmsterms)
export(is.cor_arma)
export(is.cor_brms)
export(is.cor_car)
export(is.cor_cosy)
export(is.cor_fixed)
export(is.cor_sar)
export(is.mvbrmsformula)
export(is.mvbrmsterms)
export(kfold)
export(kfold_predict)
export(lasso)
export(lf)
export(log_lik)
export(log_posterior)
export(logistic_normal)
export(logit_scaled)
export(logm1)
export(lognormal)
export(loo)
export(loo_R2)
export(loo_compare)
export(loo_epred)
export(loo_linpred)
export(loo_model_weights)
export(loo_moment_match)
export(loo_predict)
export(loo_predictive_interval)
export(loo_subsample)
export(ma)
export(make_conditions)
export(make_stancode)
export(make_standata)
export(marginal_effects)
export(marginal_smooths)
export(mcmc_plot)
export(me)
export(mi)
export(mixture)
export(mm)
export(mmc)
export(mo)
export(model_weights)
export(multinomial)
export(mvbf)
export(mvbind)
export(mvbrmsformula)
export(nchains)
export(ndraws)
export(neff_ratio)
export(negbinomial)
export(ngrps)
export(niterations)
export(nlf)
export(nsamples)
export(nuts_params)
export(nvariables)
export(opencl)
export(parnames)
export(parse_bf)
export(pasym_laplace)
export(pbeta_binomial)
export(pexgaussian)
export(pfrechet)
export(pgen_extreme_value)
export(phurdle_gamma)
export(phurdle_lognormal)
export(phurdle_negbinomial)
export(phurdle_poisson)
export(pinv_gaussian)
export(post_prob)
export(posterior_average)
export(posterior_epred)
export(posterior_interval)
export(posterior_linpred)
export(posterior_predict)
export(posterior_samples)
export(posterior_smooths)
export(posterior_summary)
export(posterior_table)
export(pp_average)
export(pp_check)
export(pp_expect)
export(pp_mixture)
export(predictive_error)
export(predictive_interval)
export(prepare_predictions)
export(prior)
export(prior_)
export(prior_draws)
export(prior_samples)
export(prior_string)
export(prior_summary)
export(pshifted_lnorm)
export(psis)
export(pskew_normal)
export(pstudent_t)
export(pvon_mises)
export(pzero_inflated_beta)
export(pzero_inflated_beta_binomial)
export(pzero_inflated_binomial)
export(pzero_inflated_negbinomial)
export(pzero_inflated_poisson)
export(qasym_laplace)
export(qfrechet)
export(qgen_extreme_value)
export(qshifted_lnorm)
export(qskew_normal)
export(qstudent_t)
export(ranef)
export(rasym_laplace)
export(rbeta_binomial)
export(rdirichlet)
export(read_csv_as_stanfit)
export(recompile_model)
export(reloo)
export(rename_pars)
export(resp_bhaz)
export(resp_cat)
export(resp_cens)
export(resp_dec)
export(resp_index)
export(resp_mi)
export(resp_rate)
export(resp_se)
export(resp_subset)
export(resp_thres)
export(resp_trials)
export(resp_trunc)
export(resp_vint)
export(resp_vreal)
export(resp_weights)
export(restructure)
export(rexgaussian)
export(rfrechet)
export(rgen_extreme_value)
export(rhat)
export(rinv_gaussian)
export(rlogistic_normal)
export(rmulti_normal)
export(rmulti_student_t)
export(rows2labels)
export(rshifted_lnorm)
export(rskew_normal)
export(rstudent_t)
export(rvon_mises)
export(rwiener)
export(s)
export(sar)
export(save_pars)
export(set_mecor)
export(set_nl)
export(set_prior)
export(set_rescor)
export(shifted_lognormal)
export(skew_normal)
export(sratio)
export(stancode)
export(standata)
export(stanplot)
export(stanvar)
export(student)
export(t2)
export(theme_black)
export(theme_default)
export(threading)
export(unstr)
export(update_adterms)
export(validate_newdata)
export(validate_prior)
export(variables)
export(von_mises)
export(waic)
export(weibull)
export(wiener)
export(zero_inflated_beta)
export(zero_inflated_beta_binomial)
export(zero_inflated_binomial)
export(zero_inflated_negbinomial)
export(zero_inflated_poisson)
export(zero_one_inflated_beta)
import(Rcpp)
import(abind)
import(ggplot2)
import(methods)
import(parallel)
import(stats)
importFrom(bayesplot,log_posterior)
importFrom(bayesplot,neff_ratio)
importFrom(bayesplot,nuts_params)
importFrom(bayesplot,pp_check)
importFrom(bayesplot,theme_default)
importFrom(bridgesampling,bayes_factor)
importFrom(bridgesampling,bridge_sampler)
importFrom(bridgesampling,post_prob)
importFrom(coda,as.mcmc)
importFrom(grDevices,devAskNewPage)
importFrom(graphics,plot)
importFrom(loo,.compute_point_estimate)
importFrom(loo,.ndraws)
importFrom(loo,.thin_draws)
importFrom(loo,is.loo)
importFrom(loo,kfold)
importFrom(loo,loo)
importFrom(loo,loo_compare)
importFrom(loo,loo_model_weights)
importFrom(loo,loo_moment_match)
importFrom(loo,loo_subsample)
importFrom(loo,psis)
importFrom(loo,waic)
importFrom(nlme,VarCorr)
importFrom(nlme,fixef)
importFrom(nlme,ranef)
importFrom(posterior,as_draws)
importFrom(posterior,as_draws_array)
importFrom(posterior,as_draws_df)
importFrom(posterior,as_draws_list)
importFrom(posterior,as_draws_matrix)
importFrom(posterior,as_draws_rvars)
importFrom(posterior,nchains)
importFrom(posterior,ndraws)
importFrom(posterior,niterations)
importFrom(posterior,nvariables)
importFrom(posterior,rhat)
importFrom(posterior,subset_draws)
importFrom(posterior,summarize_draws)
importFrom(posterior,variables)
importFrom(rlang,.data)
importFrom(rstantools,bayes_R2)
importFrom(rstantools,log_lik)
importFrom(rstantools,loo_R2)
importFrom(rstantools,loo_linpred)
importFrom(rstantools,loo_predict)
importFrom(rstantools,loo_predictive_interval)
importFrom(rstantools,nsamples)
importFrom(rstantools,posterior_epred)
importFrom(rstantools,posterior_interval)
importFrom(rstantools,posterior_linpred)
importFrom(rstantools,posterior_predict)
importFrom(rstantools,predictive_error)
importFrom(rstantools,predictive_interval)
importFrom(rstantools,prior_summary)
