% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numeric-helpers.R
\name{expp1}
\alias{expp1}
\title{Exponential function plus one.}
\usage{
expp1(x)
}
\arguments{
\item{x}{A numeric or complex vector.}
}
\description{
Computes \code{exp(x) + 1}.
}
