% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-response.R
\name{get_y}
\alias{get_y}
\title{Extract response values}
\usage{
get_y(x, resp = NULL, sort = FALSE, warn = FALSE, ...)
}
\arguments{
\item{x}{A \code{\link{brmsfit}} object.}

\item{resp}{Optional names of response variables for which to extract values.}

\item{sort}{Logical. Only relevant for time series models.
Indicating whether to return predicted values in the original
order (\code{FALSE}; default) or in the order of the
time series (\code{TRUE}).}

\item{warn}{For internal use only.}

\item{...}{Further arguments passed to \code{\link{standata}}.}
}
\value{
Returns a vector of response values for univariate models and a
  matrix of response values with one column per response variable for
  multivariate models.
}
\description{
Extract response values from a \code{\link{brmsfit}} object.
}
\keyword{internal}
